@@ -1,31 +1,39 @@
 use super::*;
 use crate::types::gossip::GossipModule;
 use ghost_actor::dependencies::tracing;
-use ghost_actor::dependencies::tracing_futures::Instrument;
 use kitsune_p2p_mdns::*;
+use kitsune_p2p_types::agent_info::AgentInfoSigned;
 use kitsune_p2p_types::codec::{rmp_decode, rmp_encode};
+use kitsune_p2p_types::dht_arc::DhtArc;
 use kitsune_p2p_types::tx2::tx2_utils::TxUrl;
 use std::collections::{HashMap, HashSet};
 use std::sync::atomic::AtomicBool;
 use url2::Url2;
 
-/// if the user specifies None or zero (0) for race_timeout_ms
-/// (david.b) this is not currently used
-const DEFAULT_RPC_MULTI_RACE_TIMEOUT_MS: u64 = 200;
+mod rpc_multi_logic;
+
+type KSpace = Arc<KitsuneSpace>;
+type KAgent = Arc<KitsuneAgent>;
+type KBasis = Arc<KitsuneBasis>;
+type WireConHnd = Tx2ConHnd<wire::Wire>;
+type Payload = Box<[u8]>;
 
 ghost_actor::ghost_chan! {
     pub(crate) chan SpaceInternal<crate::KitsuneP2pError> {
         /// List online agents that claim to be covering a basis hash
-        fn list_online_agents_for_basis_hash(space: Arc<KitsuneSpace>, from_agent: Arc<KitsuneAgent>, basis: Arc<KitsuneBasis>) -> HashSet<Arc<KitsuneAgent>>;
+        fn list_online_agents_for_basis_hash(space: KSpace, from_agent: KAgent, basis: KBasis) -> HashSet<KAgent>;
 
         /// Update / publish our agent info
         fn update_agent_info() -> ();
 
         /// Update / publish a single agent info
-        fn update_single_agent_info(agent: Arc<KitsuneAgent>) -> ();
+        fn update_single_agent_info(agent: KAgent) -> ();
 
         /// see if an agent is locally joined
-        fn is_agent_local(agent: Arc<KitsuneAgent>) -> bool;
+        fn is_agent_local(agent: KAgent) -> bool;
+
+        /// Update the arc of a local agent.
+        fn update_agent_arc(agent: KAgent, arc: DhtArc) -> ();
 
         /// Incoming Delegate Broadcast
         /// We are being requested to delegate a broadcast to our neighborhood
@@ -33,16 +41,16 @@ ghost_actor::ghost_chan! {
         /// neighbors we are responsible for.
         /// (See comments in actual method impl for more detail.)
         fn incoming_delegate_broadcast(
-            space: Arc<KitsuneSpace>,
-            basis: Arc<KitsuneBasis>,
-            to_agent: Arc<KitsuneAgent>,
+            space: KSpace,
+            basis: KBasis,
+            to_agent: KAgent,
             mod_idx: u32,
             mod_cnt: u32,
             data: crate::wire::WireData,
         ) -> ();
 
         /// Incoming Gossip
-        fn incoming_gossip(space: Arc<KitsuneSpace>, con: Tx2ConHnd<wire::Wire>, data: Box<[u8]>) -> ();
+        fn incoming_gossip(space: KSpace, con: WireConHnd, data: Payload, module_type: crate::types::gossip::GossipModuleType) -> ();
     }
 }
 
@@ -51,6 +59,8 @@ pub(crate) async fn spawn_space(
     this_addr: url2::Url2,
     ep_hnd: Tx2EpHnd<wire::Wire>,
     config: Arc<KitsuneP2pConfig>,
+    bandwidth_throttles: BandwidthThrottles,
+    parallel_notify_permit: Arc<tokio::sync::Semaphore>,
 ) -> KitsuneP2pResult<(
     ghost_actor::GhostSender<KitsuneP2p>,
     ghost_actor::GhostSender<SpaceInternal>,
@@ -77,6 +87,8 @@ pub(crate) async fn spawn_space(
         evt_send,
         ep_hnd,
         config,
+        bandwidth_throttles,
+        parallel_notify_permit,
     )));
 
     Ok((sender, i_s, evt_recv))
@@ -88,7 +100,7 @@ impl SpaceInternalHandler for Space {
     fn handle_list_online_agents_for_basis_hash(
         &mut self,
         _space: Arc<KitsuneSpace>,
-        from_agent: Arc<KitsuneAgent>,
+        _from_agent: Arc<KitsuneAgent>,
         // during short-circuit / full-sync mode,
         // we're ignoring the basis_hash and just returning everyone.
         _basis: Arc<KitsuneBasis>,
@@ -99,7 +111,7 @@ impl SpaceInternalHandler for Space {
             .evt_sender
             .query_agent_info_signed(QueryAgentInfoSignedEvt {
                 space: self.space.clone(),
-                agent: from_agent,
+                agents: None,
             });
         Ok(async move {
             for peer in all_peers_fut.await? {
@@ -115,26 +127,40 @@ impl SpaceInternalHandler for Space {
         let space = self.space.clone();
         let mut mdns_handles = self.mdns_handles.clone();
         let network_type = self.config.network_type.clone();
-        let agent_list: Vec<Arc<KitsuneAgent>> = self.local_joined_agents.iter().cloned().collect();
+        let mut agent_list = Vec::with_capacity(self.local_joined_agents.len());
+        for agent in self.local_joined_agents.iter().cloned() {
+            let arc = self.get_agent_arc(&agent);
+            agent_list.push((agent, arc));
+        }
         let bound_url = self.this_addr.clone();
         let evt_sender = self.evt_sender.clone();
         let bootstrap_service = self.config.bootstrap_service.clone();
         let expires_after = self.config.tuning_params.agent_info_expires_after_ms as u64;
+        let internal_sender = self.i_s.clone();
         Ok(async move {
             let urls = vec![bound_url.into()];
-            for agent in agent_list {
+            let mut peer_data = Vec::with_capacity(agent_list.len());
+            for (agent, arc) in agent_list {
                 let input = UpdateAgentInfoInput {
                     expires_after,
                     space: space.clone(),
                     agent,
+                    arc,
                     urls: &urls,
                     evt_sender: &evt_sender,
+                    internal_sender: &internal_sender,
                     network_type: network_type.clone(),
                     mdns_handles: &mut mdns_handles,
                     bootstrap_service: &bootstrap_service,
                 };
-                update_single_agent_info(input).await?;
+                peer_data.push(update_single_agent_info(input).await?);
             }
+            evt_sender
+                .put_agent_info_signed(PutAgentInfoSignedEvt {
+                    space: space.clone(),
+                    peer_data,
+                })
+                .await?;
             Ok(())
         }
         .boxed()
@@ -150,21 +176,32 @@ impl SpaceInternalHandler for Space {
         let network_type = self.config.network_type.clone();
         let bound_url = self.this_addr.clone();
         let evt_sender = self.evt_sender.clone();
+        let internal_sender = self.i_s.clone();
         let bootstrap_service = self.config.bootstrap_service.clone();
         let expires_after = self.config.tuning_params.agent_info_expires_after_ms as u64;
+        let arc = self.get_agent_arc(&agent);
+
         Ok(async move {
             let urls = vec![bound_url.into()];
             let input = UpdateAgentInfoInput {
                 expires_after,
                 space: space.clone(),
                 agent,
+                arc,
                 urls: &urls,
                 evt_sender: &evt_sender,
+                internal_sender: &internal_sender,
                 network_type: network_type.clone(),
                 mdns_handles: &mut mdns_handles,
                 bootstrap_service: &bootstrap_service,
             };
-            update_single_agent_info(input).await?;
+            let peer_data = vec![update_single_agent_info(input).await?];
+            evt_sender
+                .put_agent_info_signed(PutAgentInfoSignedEvt {
+                    space: space.clone(),
+                    peer_data,
+                })
+                .await?;
             Ok(())
         }
         .boxed()
@@ -179,6 +216,15 @@ impl SpaceInternalHandler for Space {
         Ok(async move { Ok(res) }.boxed().into())
     }
 
+    fn handle_update_agent_arc(
+        &mut self,
+        agent: Arc<KitsuneAgent>,
+        arc: DhtArc,
+    ) -> SpaceInternalHandlerResult<()> {
+        self.agent_arcs.insert(agent, arc);
+        Ok(async move { Ok(()) }.boxed().into())
+    }
+
     fn handle_incoming_delegate_broadcast(
         &mut self,
         space: Arc<KitsuneSpace>,
@@ -227,7 +273,6 @@ impl SpaceInternalHandler for Space {
             {
                 let ro_inner = ro_inner.clone();
                 let space = space.clone();
-                let basis = basis.clone();
                 let data = data.clone();
                 all.push(async move {
                     use discover::PeerDiscoverResult;
@@ -243,7 +288,7 @@ impl SpaceInternalHandler for Space {
                     };
 
                     // generate our broadcast payload
-                    let payload = wire::Wire::broadcast(space, basis, info.agent.clone(), data);
+                    let payload = wire::Wire::broadcast(space, info.agent.clone(), data);
 
                     // forward the data
                     if let Err(err) = con_hnd.notify(&payload, timeout).await {
@@ -265,9 +310,16 @@ impl SpaceInternalHandler for Space {
         _space: Arc<KitsuneSpace>,
         con: Tx2ConHnd<wire::Wire>,
         data: Box<[u8]>,
+        module_type: GossipModuleType,
     ) -> InternalHandlerResult<()> {
-        self.gossip_mod.incoming_gossip(con, data)?;
-        Ok(async move { Ok(()) }.boxed().into())
+        match self.gossip_mod.get(&module_type) {
+            Some(module) => module.incoming_gossip(con, data)?,
+            None => tracing::warn!(
+                "Received gossip for {:?} but this gossip module isn't running",
+                module_type
+            ),
+        }
+        unit_ok_fut()
     }
 }
 
@@ -275,32 +327,65 @@ struct UpdateAgentInfoInput<'borrow> {
     expires_after: u64,
     space: Arc<KitsuneSpace>,
     agent: Arc<KitsuneAgent>,
+    arc: DhtArc,
     urls: &'borrow Vec<TxUrl>,
     evt_sender: &'borrow futures::channel::mpsc::Sender<KitsuneP2pEvent>,
+    internal_sender: &'borrow ghost_actor::GhostSender<SpaceInternal>,
     network_type: NetworkType,
     mdns_handles: &'borrow mut HashMap<Vec<u8>, Arc<AtomicBool>>,
     bootstrap_service: &'borrow Option<Url2>,
 }
 
-async fn update_single_agent_info(input: UpdateAgentInfoInput<'_>) -> KitsuneP2pResult<()> {
+#[cfg(feature = "sharded")]
+async fn update_arc_length(
+    evt_sender: &futures::channel::mpsc::Sender<KitsuneP2pEvent>,
+    space: Arc<KitsuneSpace>,
+    arc: &mut DhtArc,
+) -> KitsuneP2pResult<()> {
+    let density = evt_sender
+        .query_peer_density(space.clone(), arc.clone())
+        .await?;
+    arc.update_length(density);
+    Ok(())
+}
+
+#[cfg(not(feature = "sharded"))]
+async fn update_arc_length(
+    _evt_sender: &futures::channel::mpsc::Sender<KitsuneP2pEvent>,
+    _space: Arc<KitsuneSpace>,
+    _arc: &mut DhtArc,
+) -> KitsuneP2pResult<()> {
+    Ok(())
+}
+
+async fn update_single_agent_info(
+    input: UpdateAgentInfoInput<'_>,
+) -> KitsuneP2pResult<AgentInfoSigned> {
     let UpdateAgentInfoInput {
         expires_after,
         space,
         agent,
+        mut arc,
         urls,
         evt_sender,
+        internal_sender,
         network_type,
         mdns_handles,
         bootstrap_service,
     } = input;
-    use kitsune_p2p_types::agent_info::AgentInfoSigned;
+
+    update_arc_length(evt_sender, space.clone(), &mut arc).await?;
+
+    // Update the agents arc through the internal sender.
+    internal_sender.update_agent_arc(agent.clone(), arc).await?;
+
     let signed_at_ms = crate::spawn::actor::bootstrap::now_once(None).await?;
     let expires_at_ms = signed_at_ms + expires_after;
 
     let agent_info_signed = AgentInfoSigned::sign(
         space.clone(),
         agent.clone(),
-        u32::MAX,
+        arc.half_length(),
         urls.clone(),
         signed_at_ms,
         expires_at_ms,
@@ -324,14 +409,6 @@ async fn update_single_agent_info(input: UpdateAgentInfoInput<'_>) -> KitsuneP2p
 
     tracing::debug!(?agent_info_signed);
 
-    evt_sender
-        .put_agent_info_signed(PutAgentInfoSignedEvt {
-            space: space.clone(),
-            agent: agent.clone(),
-            agent_info_signed: agent_info_signed.clone(),
-        })
-        .await?;
-
     // Push to the network as well
     match network_type {
         NetworkType::QuicMdns => {
@@ -357,11 +434,14 @@ async fn update_single_agent_info(input: UpdateAgentInfoInput<'_>) -> KitsuneP2p
             }
         }
         NetworkType::QuicBootstrap => {
-            crate::spawn::actor::bootstrap::put(bootstrap_service.clone(), agent_info_signed)
-                .await?;
+            crate::spawn::actor::bootstrap::put(
+                bootstrap_service.clone(),
+                agent_info_signed.clone(),
+            )
+            .await?;
         }
     }
-    Ok(())
+    Ok(agent_info_signed)
 }
 
 use ghost_actor::dependencies::must_future::MustBoxFuture;
@@ -371,7 +451,9 @@ impl ghost_actor::GhostControlHandler for Space {
             use futures::sink::SinkExt;
             // this is a curtesy, ok if fails
             let _ = self.evt_sender.close().await;
-            self.gossip_mod.close();
+            for module in self.gossip_mod.values_mut() {
+                module.close();
+            }
         }
         .boxed()
         .into()
@@ -393,14 +475,15 @@ impl KitsuneP2pHandler for Space {
         agent: Arc<KitsuneAgent>,
     ) -> KitsuneP2pHandlerResult<()> {
         self.local_joined_agents.insert(agent.clone());
-        self.gossip_mod.local_agent_join(agent.clone());
+        for module in self.gossip_mod.values() {
+            module.local_agent_join(agent.clone());
+        }
         let fut = self.i_s.update_single_agent_info(agent);
         let evt_sender = self.evt_sender.clone();
         match self.config.network_type {
             NetworkType::QuicMdns => {
                 // Listen to MDNS service that has that space as service type
                 let space_b64 = base64::encode_config(&space[..], base64::URL_SAFE_NO_PAD);
-                //println!("(MDNS) - Agent {:?} ({}) joined space {:?} ({} ; {})", agent, agent.get_bytes().len(), space, space.get_bytes().len(), dna_str.len());
                 if !self.mdns_listened_spaces.contains(&space_b64) {
                     self.mdns_listened_spaces.insert(space_b64.clone());
                     tokio::task::spawn(async move {
@@ -411,29 +494,24 @@ impl KitsuneP2pHandler for Space {
                                 Ok(response) => {
                                     tracing::trace!(msg = "Peer found via MDNS", ?response);
                                     // Decode response
-                                    let remote_agent_vec = base64::decode_config(
-                                        &response.service_name[..],
-                                        base64::URL_SAFE_NO_PAD,
-                                    )
-                                    .expect("Agent base64 decode failed");
-                                    let remote_agent = Arc::new(KitsuneAgent(remote_agent_vec));
-                                    //println!("(MDNS) - Peer found via MDNS: {:?})", *remote_agent);
                                     let maybe_agent_info_signed =
                                         rmp_decode(&mut &*response.buffer);
                                     if let Err(e) = maybe_agent_info_signed {
                                         tracing::error!(msg = "Failed to decode MDNS peer", ?e);
                                         continue;
                                     }
-                                    let remote_agent_info_signed = maybe_agent_info_signed.unwrap();
-                                    //println!("(MDNS) - Found agent_info_signed: {:?})", remote_agent_info_signed);
-                                    // Add to local storage
-                                    let _result = evt_sender
-                                        .put_agent_info_signed(PutAgentInfoSignedEvt {
-                                            space: space.clone(),
-                                            agent: remote_agent,
-                                            agent_info_signed: remote_agent_info_signed,
-                                        })
-                                        .await;
+                                    if let Ok(remote_agent_info_signed) = maybe_agent_info_signed {
+                                        // Add to local storage
+                                        if let Err(e) = evt_sender
+                                            .put_agent_info_signed(PutAgentInfoSignedEvt {
+                                                space: space.clone(),
+                                                peer_data: vec![remote_agent_info_signed],
+                                            })
+                                            .await
+                                        {
+                                            tracing::error!(msg = "Failed to store MDNS peer", ?e);
+                                        }
+                                    }
                                 }
                                 Err(e) => {
                                     tracing::error!(msg = "Failed to get peers from MDNS", ?e);
@@ -458,7 +536,10 @@ impl KitsuneP2pHandler for Space {
         agent: Arc<KitsuneAgent>,
     ) -> KitsuneP2pHandlerResult<()> {
         self.local_joined_agents.remove(&agent);
-        self.gossip_mod.local_agent_leave(agent.clone());
+        self.agent_arcs.remove(&agent);
+        for module in self.gossip_mod.values() {
+            module.local_agent_leave(agent.clone());
+        }
         self.publish_leave_agent_info(agent)
     }
 
@@ -513,43 +594,14 @@ impl KitsuneP2pHandler for Space {
 
     fn handle_rpc_multi(
         &mut self,
-        mut input: actor::RpcMulti,
+        input: actor::RpcMulti,
     ) -> KitsuneP2pHandlerResult<Vec<actor::RpcMultiResponse>> {
-        // if the user doesn't care about remote_agent_count, apply default
-        match input.remote_agent_count {
-            None | Some(0) => {
-                input.remote_agent_count = Some(
-                    self.config
-                        .tuning_params
-                        .default_rpc_multi_remote_agent_count as u8,
-                );
-            }
-            _ => {}
-        }
-
-        // if the user doesn't care about timeout_ms, apply default
-        match input.timeout_ms {
-            None | Some(0) => {
-                input.timeout_ms =
-                    Some(self.config.tuning_params.default_rpc_multi_timeout_ms as u64);
-            }
-            _ => {}
-        }
-
-        // if the user doesn't care about race_timeout_ms, apply default
-        match input.race_timeout_ms {
-            None | Some(0) => {
-                input.race_timeout_ms = Some(DEFAULT_RPC_MULTI_RACE_TIMEOUT_MS);
-            }
-            _ => {}
-        }
-
-        // race timeout > timeout is nonesense
-        if input.as_race && input.race_timeout_ms.unwrap() > input.timeout_ms.unwrap() {
-            input.race_timeout_ms = Some(input.timeout_ms.unwrap());
-        }
-
-        self.handle_rpc_multi_inner(input)
+        let fut = rpc_multi_logic::handle_rpc_multi(
+            input,
+            self.ro_inner.clone(),
+            self.local_joined_agents.clone(),
+        );
+        Ok(async move { fut.await }.boxed().into())
     }
 
     fn handle_broadcast(
@@ -590,6 +642,12 @@ impl KitsuneP2pHandler for Space {
             // we won't get notified if we are unable to publish to anyone.
             // Also, if conductor spams us with publishes, we could fill
             // the memory with publish tasks.
+            let task_permit = ro_inner
+                .parallel_notify_permit
+                .clone()
+                .acquire_owned()
+                .await
+                .ok();
             tokio::task::spawn(async move {
                 let cover_nodes = discover_fut.await?;
                 if cover_nodes.is_empty() {
@@ -663,6 +721,7 @@ impl KitsuneP2pHandler for Space {
 
                 futures::future::join_all(all).await;
 
+                drop(task_permit);
                 KitsuneP2pResult::Ok(())
             });
 
@@ -671,6 +730,120 @@ impl KitsuneP2pHandler for Space {
         .boxed()
         .into())
     }
+
+    fn handle_targeted_broadcast(
+        &mut self,
+        space: Arc<KitsuneSpace>,
+        from_agent: Arc<KitsuneAgent>,
+        agents: Vec<Arc<KitsuneAgent>>,
+        timeout: KitsuneTimeout,
+        payload: Vec<u8>,
+    ) -> KitsuneP2pHandlerResult<()> {
+        let evt_sender = self.evt_sender.clone();
+        let ro_inner = self.ro_inner.clone();
+        let concurrent_limit_per_thread = self.config.tuning_params.concurrent_limit_per_thread;
+        Ok(async move {
+            // limit spawns with semaphore.
+            let task_permit = ro_inner
+                .parallel_notify_permit
+                .clone()
+                .acquire_owned()
+                .await
+                .ok();
+            tokio::task::spawn(async move {
+                let mut futures = Vec::with_capacity(agents.len());
+                for agent in agents {
+                    let discover_fut = async {
+                        let result = discover::search_and_discover_peer_connect(
+                            ro_inner.clone(),
+                            agent.clone(),
+                            timeout,
+                        )
+                        .await;
+                        (result, agent)
+                    };
+                    futures.push(discover_fut);
+                }
+                let futures = futures::stream::iter(futures);
+                let futures = futures.buffer_unordered(concurrent_limit_per_thread);
+                futures
+                    .for_each_concurrent(concurrent_limit_per_thread, |(discover_result, agent)| {
+                        match discover_result {
+                            discover::PeerDiscoverResult::OkShortcut => {
+                                // reflect this request locally
+                                evt_sender
+                                    .notify(
+                                        space.clone(),
+                                        agent,
+                                        from_agent.clone(),
+                                        payload.clone(),
+                                    )
+                                    .map(|r| {
+                                        if let Err(e) = r {
+                                            tracing::error!(
+                                                "Failed to broadcast to local agent because: {:?}",
+                                                e
+                                            )
+                                        }
+                                    })
+                                    .boxed()
+                            }
+                            discover::PeerDiscoverResult::OkRemote { con_hnd, .. } => {
+                                let payload = wire::Wire::broadcast(
+                                    space.clone(),
+                                    agent,
+                                    payload.clone().into(),
+                                );
+                                con_hnd
+                                    .notify(&payload, timeout)
+                                    .map(|r| {
+                                        if let Err(e) = r {
+                                            tracing::info!(
+                                                "Failed to broadcast to remote agent because: {:?}",
+                                                e
+                                            )
+                                        }
+                                    })
+                                    .boxed()
+                            }
+                            discover::PeerDiscoverResult::Err(e) => async move {
+                                tracing::info!(
+                                    "Failed to discover connection for {:?} because: {:?}",
+                                    agent,
+                                    e
+                                );
+                            }
+                            .boxed(),
+                        }
+                    })
+                    .await;
+                drop(task_permit)
+            });
+            Ok(())
+        }
+        .boxed()
+        .into())
+    }
+
+    fn handle_new_integrated_data(&mut self, _: KSpace) -> InternalHandlerResult<()> {
+        for module in self.gossip_mod.values() {
+            module.new_integrated_data();
+        }
+        unit_ok_fut()
+    }
+
+    fn handle_authority_for_hash(
+        &mut self,
+        _space: Arc<KitsuneSpace>,
+        agent: Arc<KitsuneAgent>,
+        basis: Arc<KitsuneBasis>,
+    ) -> KitsuneP2pHandlerResult<bool> {
+        let r = match self.agent_arcs.get(&agent) {
+            Some(agent_arc) => agent_arc.contains(basis.get_loc()),
+            None => false,
+        };
+        Ok(async move { Ok(r) }.boxed().into())
+    }
 }
 
 pub(crate) struct SpaceReadOnlyInner {
@@ -682,6 +855,7 @@ pub(crate) struct SpaceReadOnlyInner {
     pub(crate) ep_hnd: Tx2EpHnd<wire::Wire>,
     #[allow(dead_code)]
     pub(crate) config: Arc<KitsuneP2pConfig>,
+    pub(crate) parallel_notify_permit: Arc<tokio::sync::Semaphore>,
 }
 
 /// A Kitsune P2p Node can track multiple "spaces" -- Non-interacting namespaced
@@ -693,14 +867,16 @@ pub(crate) struct Space {
     pub(crate) i_s: ghost_actor::GhostSender<SpaceInternal>,
     pub(crate) evt_sender: futures::channel::mpsc::Sender<KitsuneP2pEvent>,
     pub(crate) local_joined_agents: HashSet<Arc<KitsuneAgent>>,
+    pub(crate) agent_arcs: HashMap<Arc<KitsuneAgent>, DhtArc>,
     pub(crate) config: Arc<KitsuneP2pConfig>,
     mdns_handles: HashMap<Vec<u8>, Arc<AtomicBool>>,
     mdns_listened_spaces: HashSet<String>,
-    gossip_mod: GossipModule,
+    gossip_mod: HashMap<GossipModuleType, GossipModule>,
 }
 
 impl Space {
     /// space constructor
+    #[allow(clippy::too_many_arguments)]
     pub fn new(
         space: Arc<KitsuneSpace>,
         this_addr: url2::Url2,
@@ -708,21 +884,47 @@ impl Space {
         evt_sender: futures::channel::mpsc::Sender<KitsuneP2pEvent>,
         ep_hnd: Tx2EpHnd<wire::Wire>,
         config: Arc<KitsuneP2pConfig>,
+        bandwidth_throttles: BandwidthThrottles,
+        parallel_notify_permit: Arc<tokio::sync::Semaphore>,
     ) -> Self {
-        let gossip_mod_fact = if &config.tuning_params.gossip_strategy == "simple-bloom" {
-            crate::gossip::simple_bloom::factory()
-        } else {
-            panic!(
-                "unknown gossip strategy: {}",
-                config.tuning_params.gossip_strategy
-            );
-        };
-        let gossip_mod = gossip_mod_fact.spawn_gossip_task(
-            config.tuning_params.clone(),
-            space.clone(),
-            ep_hnd.clone(),
-            evt_sender.clone(),
-        );
+        let gossip_mod = config
+            .tuning_params
+            .gossip_strategy
+            .split(',')
+            .flat_map(|module| match module {
+                "simple-bloom" => vec![(
+                    GossipModuleType::Simple,
+                    crate::gossip::simple_bloom::factory(),
+                )],
+                "sharded-gossip" => vec![
+                    (
+                        GossipModuleType::ShardedRecent,
+                        crate::gossip::sharded_gossip::recent_factory(bandwidth_throttles.recent()),
+                    ),
+                    (
+                        GossipModuleType::ShardedHistorical,
+                        crate::gossip::sharded_gossip::historical_factory(
+                            bandwidth_throttles.historical(),
+                        ),
+                    ),
+                ],
+                "none" => vec![],
+                _ => {
+                    panic!("unknown gossip strategy: {}", module);
+                }
+            })
+            .map(|(module, factory)| {
+                (
+                    module,
+                    factory.spawn_gossip_task(
+                        config.tuning_params.clone(),
+                        space.clone(),
+                        ep_hnd.clone(),
+                        evt_sender.clone(),
+                    ),
+                )
+            })
+            .collect();
 
         let i_s_c = i_s.clone();
         tokio::task::spawn(async move {
@@ -773,31 +975,28 @@ impl Space {
                             if !i_s_c.ghost_actor_is_active() {
                                 break;
                             }
+                            let mut peer_data = Vec::with_capacity(list.len());
                             for item in list {
                                 // TODO - someday some validation here
-                                let agent = item.agent.clone();
-                                match i_s_c.is_agent_local(agent.clone()).await {
+                                match i_s_c.is_agent_local(item.agent.clone()).await {
                                     Err(err) => tracing::error!(?err),
                                     Ok(is_local) => {
                                         if !is_local {
                                             // we got a result - let's add it to our store for the future
-                                            if let Err(err) = evt_s_c
-                                                .put_agent_info_signed(PutAgentInfoSignedEvt {
-                                                    space: space_c.clone(),
-                                                    agent,
-                                                    agent_info_signed: item.clone(),
-                                                })
-                                                .await
-                                            {
-                                                tracing::error!(
-                                                    ?err,
-                                                    "error storing bootstrap agent_info"
-                                                );
-                                            }
+                                            peer_data.push(item);
                                         }
                                     }
                                 }
                             }
+                            if let Err(err) = evt_s_c
+                                .put_agent_info_signed(PutAgentInfoSignedEvt {
+                                    space: space_c.clone(),
+                                    peer_data,
+                                })
+                                .await
+                            {
+                                tracing::error!(?err, "error storing bootstrap agent_info");
+                            }
                         }
                     }
                 }
@@ -812,6 +1011,7 @@ impl Space {
             evt_sender: evt_sender.clone(),
             ep_hnd,
             config: config.clone(),
+            parallel_notify_permit,
         });
 
         Self {
@@ -821,6 +1021,7 @@ impl Space {
             i_s,
             evt_sender,
             local_joined_agents: HashSet::new(),
+            agent_arcs: HashMap::new(),
             config,
             mdns_handles: HashMap::new(),
             mdns_listened_spaces: HashSet::new(),
@@ -840,7 +1041,6 @@ impl Space {
         Ok(async move {
             let signed_at_ms = crate::spawn::actor::bootstrap::now_once(None).await?;
             let expires_at_ms = signed_at_ms + expires_after;
-            use kitsune_p2p_types::agent_info::AgentInfoSigned;
             let agent_info_signed = AgentInfoSigned::sign(
                 space.clone(),
                 agent.clone(),
@@ -871,8 +1071,7 @@ impl Space {
             evt_sender
                 .put_agent_info_signed(PutAgentInfoSignedEvt {
                     space: space.clone(),
-                    agent: agent.clone(),
-                    agent_info_signed: agent_info_signed.clone(),
+                    peer_data: vec![agent_info_signed.clone()],
                 })
                 .await?;
 
@@ -894,101 +1093,13 @@ impl Space {
         .into())
     }
 
-    /// actual logic for handle_rpc_multi ...
-    /// the top-level handler may or may not spawn a task for this
-    #[tracing::instrument(skip(self, input))]
-    fn handle_rpc_multi_inner(
-        &mut self,
-        input: actor::RpcMulti,
-    ) -> KitsuneP2pHandlerResult<Vec<actor::RpcMultiResponse>> {
-        let actor::RpcMulti {
-            space,
-            from_agent,
-            //basis,
-            //remote_agent_count,
-            //timeout_ms,
-            //as_race,
-            //race_timeout_ms,
-            payload,
-            ..
-        } = input;
-
-        // TODO - FIXME - david.b - removing the parts of this that
-        // actually make remote requests. We can get this data locally
-        // while we are still full sync after gossip, and the timeouts
-        // are not structured correctly.
-        //
-        // Better to re-write as part of sharding.
-
-        //let remote_agent_count = remote_agent_count.unwrap();
-        //let timeout_ms = timeout_ms.unwrap();
-        //let stage_1_timeout_ms = timeout_ms / 2;
-
-        // as an optimization - request to all local joins
-        // but don't count that toward our request total
-        let local_all = self
-            .local_joined_agents
-            .iter()
-            .map(|agent| {
-                let agent = agent.clone();
-                self.evt_sender
-                    .call(
-                        space.clone(),
-                        agent.clone(),
-                        from_agent.clone(),
-                        payload.clone(),
-                    )
-                    .then(|r| async move { (r, agent) })
-            })
-            .collect::<Vec<_>>();
-
-        /*
-        let remote_fut = discover::message_neighborhood(
-            self,
-            from_agent.clone(),
-            remote_agent_count,
-            stage_1_timeout_ms,
-            timeout_ms,
-            basis,
-            wire::Wire::call(
-                space.clone(),
-                from_agent.clone(),
-                from_agent,
-                payload.clone().into(),
-            ),
-            |a, w| match w {
-                wire::Wire::CallResp(c) => Ok(actor::RpcMultiResponse {
-                    agent: a,
-                    response: c.data.into(),
-                }),
-                _ => Err(()),
-            },
-        )
-        .instrument(tracing::debug_span!("message_neighborhood", payload = ?payload.iter().take(5).collect::<Vec<_>>()));
-        */
-
-        Ok(async move {
-            let out: Vec<actor::RpcMultiResponse> = futures::future::join_all(local_all)
-                .await
-                .into_iter()
-                .filter_map(|(r, a)| {
-                    if let Ok(r) = r {
-                        Some(actor::RpcMultiResponse {
-                            agent: a,
-                            response: r,
-                        })
-                    } else {
-                        None
-                    }
-                })
-                .collect();
-
-            //out.append(&mut remote_fut.await);
-
-            Ok(out)
-        }
-        .instrument(tracing::debug_span!("multi_inner"))
-        .boxed()
-        .into())
+    /// Get the existing agent storage arc or create a new one.
+    fn get_agent_arc(&self, agent: &Arc<KitsuneAgent>) -> DhtArc {
+        // TODO: We are simply setting the initial arc to full.
+        // In the future we may want to do something more intelligent.
+        self.agent_arcs
+            .get(agent)
+            .cloned()
+            .unwrap_or_else(|| DhtArc::full(agent.get_loc()))
     }
 }
